import PhoneModes
import datetime
import LexerExpect
import Lexer
import MessageBuilder
class CommandProtocol:

    """While this class successfully abstract away many of the details about the D600, there's still plenty of
        details that propagate up, e.g.:
            * The meaning of the fields.
            * That Events are modeless, Messages have different modes but the same interface, while Contacts have different
             interfaces per mode.
            """
    
    def __init__(self, phoneConnection):
        self.pc = phoneConnection
        self.le = LexerExpect.LexerExpect(Lexer.D600PhoneLexer(phoneConnection))
        self.pm = PhoneModes.PhoneModes(self)

    def close():
        self.pc.close()
        self.pc = None
        self.le = None
        self.pm = None

    # ---- Mode Setting ---- (These are used by, and don't touch, the PhoneModes object)

    def setEcho(self, echoOn):
        if echoOn:
            assert not echoOn, "ECHO ON won't work with rest of system, as it will affect responses."
            echoMode = 'On'
        else:
            echoMode = 'Off'
        command, heardYou = MessageBuilder.buildChangeModeCommand('Echo',echoMode)        
        self.pc.write(command)
        if self.le.nextTokenIsOK():
            # Echo was already off
            self.le.receiveOK()
        else:
            self.le.receiveConstantWord(heardYou)
            self.le.receiveOK()            

    def setPhoneBookMode(self, mode):
        # Assumes echo off
        assert mode in MessageBuilder.legalPhoneBookModes
        command, heardYou = MessageBuilder.buildChangeModeCommand('PhoneBook', mode)
        self.pc.write(command)
        assert heardYou == ''
        self.le.receiveOK()


    def setMessageBookMode(self, mode):
        # Assumes echo off        
        assert mode in MessageBuilder.legalMessageBookModes
        command, heardYou = MessageBuilder.buildChangeModeCommand('MessageBook', mode)
        self.pc.write(command)
        self.le.receiveConstantWord(heardYou)

        # Discard summary information
        self.le.receiveNumber(True)
        self.le.receiveNumber(True)
        self.le.receiveNumber(True)
        self.le.receiveNumber(True)
        self.le.receiveNumber(True)
        self.le.receiveNumber(False)
        
        self.le.receiveOK()        


    # ---- Phone Books and Contacts

    def getPhoneBookSummary(self, mode):
        self.pm.ensurePhoneBookMode(mode)
        command, heardYou = MessageBuilder.buildGeneralCommand('RequestPhoneBookSummary')
        self.pc.write(command)
        
        self.le.receiveConstantWord(heardYou)
        # Abstraction leak! The heardYou code doesn't handle multiple tokens, like here.
        self.le.receiveConstantQuotedString(MessageBuilder.phoneBookModeDefinitions[mode],True)

        result = {}        
        result["count"] = self.le.receiveNumber(True)
        result["capacity"] = self.le.receiveNumber(False)        

        self.le.receiveOK()

        return result

    def getPhoneMemoryContact(self, number):
        
        assert isinstance(number, int)
        self.pm.ensurePhoneBookMode('PhoneMemoryContacts')
        
        (command, heardYou) = MessageBuilder.buildGeneralCommand('RequestContact',position = number)
        self.pc.write(command)        

        result = {}
        
        if self.le.nextTokenIsOK() or self.le.nextTokenIsError():
            # Got immediate OK result, meaning no such contact
            self.le.receiveOK()
            return None

        self.le.receiveConstantWord(heardYou)
        result['contactId'] = self.le.receiveNumber(True)

        result['cellNumber'] = self.le.receiveQuotedString(True)
        result['cellNumberAddressType'] = self.le.receiveNumber(True)

        result['lastName'] = self.le.receiveSTXQuotedString(True)
        result['firstName'] = self.le.receiveSTXQuotedString(True)

        result['homeNumber'] = self.le.receiveQuotedString(True)
        result['homeNumberAddressType'] = self.le.receiveNumber(True)
        
        result['officeNumber'] = self.le.receiveQuotedString(True)
        result['officeNumberAddressType'] = self.le.receiveNumber(True)
        
        result['faxNumber'] = self.le.receiveQuotedString(True)
        result['faxNumberAddressType'] = self.le.receiveNumber(True)

        result['otherNumber'] = self.le.receiveQuotedString(True)
        result['otherNumberAddressType'] = self.le.receiveNumber(True)        

        result['emailAddress'] = self.le.receiveQuotedString(True)

        result['unknown1'] = self.le.receiveConstantQuotedString('', True)

        result['notes'] = self.le.receiveSTXQuotedString(True)

        result['groupsBitfield'] = self.le.receiveNumber(True)
        result['specialRingToneTypeId'] = self.le.receiveNumber(True)
        result['builtInRingToneId'] = self.le.receiveNumber(True)
        result['builtInAvatarId'] = self.le.receiveNumber(True)
        result['specialAvatarTypeId'] = self.le.receiveNumber(True)
        result['specialRingToneFilename'] = self.le.receiveQuotedString(True)
        result['specialAvatarFilename'] = self.le.receiveQuotedString(False)

        self.le.receiveOK()

        return result

    def getSIMMemoryContact(self, number):
        assert isinstance(number, int)
        self.pm.ensurePhoneBookMode('SIMCardContacts')
        
        (command, heardYou) = MessageBuilder.buildGeneralCommand('RequestContact',position = number)
        self.pc.write(command)        

        if self.le.nextTokenIsOK():
            # Got immediate OK result, meaning no such contact
            self.le.receiveOK()
            return None

        heardYouValue = self.le.receiveWord(False)

        if heardYouValue <> heardYou:
            # Got an oddly formatted 'not present' error
            # Bug in firmware or deliberate?
            if heardYouValue <> '+CME':
                raise LexerExpect.ParseError(heardYouValue,heardYou + ' or +CMS')
            self.le.receiveConstantWord('ERROR:') # Note colon means it isn't an ERROR token.
            self.le.receiveNumber()
            return None
        
        result = {}

        result['locationInSIM'] = self.le.receiveNumber(True)

        result['phoneNumber'] = self.le.receiveQuotedString(True)
        
        # 129. Meaning: The SMSC address is formatted using the typical ISDN / telephony numbering plan (ITU E.164/E.163) but it is not sure whether the SMSC address is an international number, a national number or a number of other types. Example addresses: "85291234567" and "91234567".
        # 145. Meaning: The SMSC address is formatted using the typical ISDN / telephony numbering plan (ITU E.164/E.163) and it is an international number. Example address: "+85291234567".
        # 255. Seems to mean default or missing.
        result['phoneNumberAddressType'] = self.le.receiveNumber(True) # Unknown

        result['name'] = self.le.receiveSTXQuotedString(False)

        self.le.receiveOK()

        return result
    
    def writePhoneMemoryContact(self, fields):
        self.pm.ensurePhoneBookMode('PhoneMemoryContacts')
        (command, heardYou)= MessageBuilder.buildPhoneMemoryContactWriteCommand(fields, 'Abort')
        
        self.pc.write(command)

        self.le.receiveConstantWord(heardYou)
        # Abstraction leak! The heardYou code doesn't handle multiple tokens, like here.
        contactId = self.le.receiveNumber(False)
        self.le.receiveOK()

        return contactId

    def writeSIMCardContact(self, fields):
        self.pm.ensurePhoneBookMode('SIMCardContacts')
        (command, heardYou)= MessageBuilder.buildSIMCardContactWriteCommand(fields, 'Abort')
        
        self.pc.write(command)
        if self.le.nextTokenIsOK():
            # Sometimes you get an immediate result. Odd.
            self.le.receiveOK()
            return fields.get("contactId")
        
        self.le.receiveConstantWord(heardYou)
        # Abstraction leak! The heardYou code doesn't handle multiple tokens, like here.
        contactId = self.le.receiveNumber(False)
        self.le.receiveOK()

        return contactId

    def deletePhoneMemoryContact(self, contactId):
        self.pm.ensurePhoneBookMode('PhoneMemoryContacts')
        (command, heardYou)= MessageBuilder.buildPhoneMemoryContactDeleteCommand(contactId, 'Abort')
        
        self.pc.write(command)        
        self.le.receiveConstantWord(heardYou)
        # Abstraction leak! The heardYou code doesn't handle multiple tokens, like here.
        self.le.receiveConstantWord(str(contactId),False)
        self.le.receiveOK()

        return contactId

    def deleteSIMCardContact(self, contactId):
        self.pm.ensurePhoneBookMode('SIMCardContacts')
        (command, heardYou)= MessageBuilder.buildSIMCardContactDeleteCommand(contactId, 'Abort')
        
        self.pc.write(command)        

        if self.le.nextTokenIsError():
            # No such contact
            self.le.receiveError()
            return None

        self.le.receiveConstantWord(heardYou)
        # Abstraction leak! The heardYou code doesn't handle multiple tokens, like here.
        self.le.receiveConstantWord(str(contactId),False)
        self.le.receiveOK()

        return contactId
    

    # ---- Message Book and Messages

    def getMessageBookSummary(self, mode):
        # Slightly innefficient, because the same information is thrown out when we enter
        # the MessageBook Mode.
        self.pm.ensureMessageBookMode(mode)
        command, heardYou = MessageBuilder.buildGeneralCommand('RequestMessageBookSummary')
        self.pc.write(command)
        self.le.receiveConstantWord(heardYou)
        result = {}

        result["mode"] = self.le.receiveQuotedString(True)
        result["count"] = self.le.receiveNumber(True)
        result["capacity"] = self.le.receiveNumber(True)

        # Data appears in triplicate? Discard rest and hope it doesn't contain any info I don't understand.
        
        self.le.receiveQuotedString(True)
        self.le.receiveNumber(True)
        self.le.receiveNumber(True)

        self.le.receiveQuotedString(True)
        self.le.receiveNumber(True)
        self.le.receiveNumber(False)

        self.le.receiveOK()        

        return result

    def getMessage(self, number, mode = None):
        assert isinstance(number, int)
        
        self.pm.ensureEchoOff()
        if mode:
            self.pm.ensureMessageBookMode(mode)
            # Otherwise you stay with the current mode.
        
        (command, heardYou) = MessageBuilder.buildGeneralCommand('RequestMessage',position = number)
        self.pc.write(command)        

        if self.le.nextTokenIsError():
            # No such message
            self.le.receiveError()
            return None

        heardYouValue = self.le.receiveWord()

        if heardYouValue <> heardYou:
            # Got an oddly formatted 'not present' error
            # Bug in firmware or deliberate?
            if heardYouValue <> '+CMS':
                raise ParseError(heardYouValue,heardYou + ' or +CMS')
            self.le.receiveConstantWord('ERROR:') # Note colon means it isn't an ERROR token.
            self.le.receiveNumber()
            return None

        # Some of the unknown fields are going to be:
        # validity period, reply path? delivery status? message Type
        result = {}


        result['state'] = self.le.receiveQuotedString(True)
        result['fromNumber'] = self.le.receiveQuotedString(True)
        result['unknownBinary'] = self.le.receiveQuotedString(True) # Junk

        # optional field!?
        if self.le.nextTokenIsQuotedString():
            result['date'] = self.le.receiveQuotedString(True)
        else:
            result['date'] = None
            
        result['UnknownInt1'] = self.le.receiveNumber(True)
        result['UnknownInt2'] = self.le.receiveNumber(True)
        result['UnknownInt3'] = self.le.receiveNumber(True)
        result['UnknownInt4'] = self.le.receiveNumber(True)

        result['messageCentrePhoneNumber'] = self.le.receiveQuotedString(True)

        # Is UnknownInt 5 an AddressType?
        result['UnknownInt5'] = self.le.receiveNumber(True)
        result['text'] = self.le.receivePascalString(True)
        result['UnknownInt6'] = self.le.receiveNumber(False)

        self.le.receiveOK()

        return result

    def deleteMessage(self, eventId):
        (command, heardYou)= MessageBuilder.buildMessageDeleteCommand(eventId, 'Abort')
        
        self.pc.write(command)
        assert(heardYou == None)
        self.le.receiveOK()

    # ---- Appointment Book and Events

    def getAppointmentBookSummary(self):
        self.pm.ensureEchoOff()
        command, heardYou = MessageBuilder.buildGeneralCommand('RequestAppointmentBookSummary')
        self.pc.write(command)
        
        self.le.receiveConstantWord(heardYou)

        result = {}        
        result["count"] = self.le.receiveNumber(True)
        result["capacity"] = self.le.receiveNumber(True)
        result["maxSubjectLength"] = self.le.receiveNumber(True)
        result["maxTextLength"] = self.le.receiveNumber(True)
        result["maxSubjectLength2"] = self.le.receiveNumber(False)

        self.le.receiveOK()

        return result

    def getEvent(self, number):
               
        assert isinstance(number, int)
        
        self.pm.ensureEchoOff()
        
        (command, heardYou) = MessageBuilder.buildGeneralCommand('RequestEvent',position = number)
        self.pc.write(command)

        if self.le.nextTokenIsError():
            # No such message
            self.le.receiveError()
            print "oops! Got an error in the event ",number,command
            return None

        heardYouValue = self.le.receiveWord()

        if heardYouValue <> heardYou:
            # Got an oddly formatted 'not present' error
            if heardYouValue <> '+CME':
                raise LexerExpect.ParseError(heardYouValue,heardYou + ' or +CMS')
            self.le.receiveConstantWord('ERROR:') # Note colon means it isn't an ERROR token.
            self.le.receiveNumber()
            return None
        result = {}

        result['eventId'] = self.le.receiveNumber(True)

        result['eventType'] = self.le.receiveNumber(True)
        # 1 = Schedule, 2 = Anniversary, 3 = To do,4 = Miscellaneous
        
        # Subject - Schedule and Miscellaneous only
        if self.le.nextTokenIsComma():
            self.le.receiveComma()
        else:
            result['subject'] = self.le.receiveSTXQuotedString(True)

        if self.le.nextTokenIsComma():
            self.le.receiveComma()
        else:
            result['text'] = self.le.receiveSTXQuotedString(True)

        if self.le.nextTokenIsComma():
            self.le.receiveComma()
        else:
            result['startDay'] = self.le.receiveNumber(True)
        if self.le.nextTokenIsComma():
            self.le.receiveComma()
        else:
            result['startMonth'] = self.le.receiveNumber(True)
        if self.le.nextTokenIsComma():
            self.le.receiveComma()
        else:
            result['startYear'] = self.le.receiveNumber(True)
            
        # Start Hour and Minute - Schedule, Anniversary and Miscellaneous only
        if self.le.nextTokenIsComma():
            self.le.receiveComma()
        else:
            result['startHour'] = self.le.receiveNumber(True)
        if self.le.nextTokenIsComma():
            self.le.receiveComma()
        else:
            result['startMinute'] = self.le.receiveNumber(True)        

        # End Date - Schedule, To do and Miscellaneous only
        if self.le.nextTokenIsComma():
            self.le.receiveComma()
        else:
            result['endDay'] = self.le.receiveNumber(True)
        if self.le.nextTokenIsComma():
            self.le.receiveComma()
        else:
            result['endMonth'] = self.le.receiveNumber(True)
        if self.le.nextTokenIsComma():
            self.le.receiveComma()
        else:
            result['endYear'] = self.le.receiveNumber(True)
        

        # End Hour and Minute - Schedule and Miscellaneous only        
        if self.le.nextTokenIsComma():
            self.le.receiveComma()
        else:
            result['endHour'] = self.le.receiveNumber(True)
        
        if self.le.nextTokenIsComma():
            self.le.receiveComma()
        else:
            result['endMinute'] = self.le.receiveNumber(True)

        # Location - Schedule only
        if self.le.nextTokenIsComma():
                self.le.receiveComma()
        else:
            result['location'] = self.le.receiveSTXQuotedString(True)

        # Alarm Fields - Schedule, Anniversary only
        if self.le.nextTokenIsComma():
                self.le.receiveComma()
        else:
            result['alarmOn'] = self.le.receiveNumber(True)
            # 1 = True, 0 = False - this record does not contain Alarm Tone info.
        if self.le.nextTokenIsComma():
                self.le.receiveComma()
        else:
            result['alarmBeforeUnits'] = self.le.receiveNumber(True)
            # 1 = Minute, 2 = Hour (not tested!), 3 = Day, 4 = Week
        if self.le.nextTokenIsComma():
                self.le.receiveComma()
        else:
            result['alarmBeforeCount'] = self.le.receiveNumber(True)
            # how many alarmBeforeUnits before the start time to alarm.

        # Recurrence Fields - Schedule and Anniversary only        
        if self.le.nextTokenIsComma():
                self.le.receiveComma()
        else:
            result['recurrence'] = self.le.receiveNumber(True)
            # 0 = No Recurrence, 1 = daily, 2 = weekly, 3 = monthly, 4 = yearly
            # Anniversary may be restricted to 4 (untested)

        # Priority - To Do only        
        if self.le.nextTokenIsComma():
                self.le.receiveComma()
        else:
            result['priority'] = self.le.receiveNumber(True)
            # 1 = High, 2 = Normal, 3 = Low
            
        # Complete - To Do only        
        if self.le.nextTokenIsComma():
                self.le.receiveComma()
        else:
            result['complete'] = self.le.receiveNumber(True)
            # 1 = Complete, 2 = Incomplete
            
        # Recurrence End Date - Schedule only
        if self.le.nextTokenIsComma():
                self.le.receiveComma()
        else:
            result['recurrenceEndDay'] = self.le.receiveNumber(True)
        if self.le.nextTokenIsComma():
                self.le.receiveComma()
        else:
            result['recurrenceEndMonth'] = self.le.receiveNumber(True)
        if not self.le.nextTokenIsOK():  # Note: Different to others!
            result['recurrenceEndYear'] = self.le.receiveNumber(False)

        self.le.receiveOK()

        return result

    def writeEvent(self, fields):
        self.pm.ensureEchoOff()
        (command, heardYou)= MessageBuilder.buildEventWriteCommand(fields, 'Abort')
        
        self.pc.write(command)
        
        self.le.receiveConstantWord(heardYou)
        positionId = self.le.receiveNumber(False)
        self.le.receiveComma()
        eventId = self.le.receiveNumber(False)        
        self.le.receiveOK()

        return (positionId, eventId)

    def deleteEvent(self, eventId):
        (command, heardYou)= MessageBuilder.buildEventDeleteCommand(eventId, 'Abort')
        
        self.pc.write(command)
        assert(heardYou == None)
        self.le.receiveOK()

        return eventId

    # --- Directories and Files 

    def readFolderList(self, path):
        self.pm.ensureEchoOff()        
        (command, heardYou)= MessageBuilder.buildReadFoldersCommand(path, 'Abort')
        self.pc.write(command)        
        resultList = []
        
        # +FSDI: 0,"Images",0,"",4444,0,0,"",47,0,0,"","","",2,1,""

        while not self.le.nextTokenIsOK():
            thisEntry = {}
            self.le.receiveConstantWord(heardYou)
            thisEntry["unknown1"] = self.le.receiveNumber(True)
            thisEntry["folderName"] = self.le.receiveQuotedString(True)
            thisEntry["unknown2"] = self.le.receiveNumber(True)

            thisEntry["unknown3"] = self.le.receiveQuotedString(True)
            thisEntry["unknown4"] = self.le.receiveNumber(True)
            thisEntry["unknown5"] = self.le.receiveNumber(True)
            thisEntry["unknown6"] = self.le.receiveNumber(True)

            thisEntry["unknown7"] = self.le.receiveQuotedString(True)
            thisEntry["unknown8"] = self.le.receiveNumber(True)
            thisEntry["unknown9"] = self.le.receiveNumber(True)
            thisEntry["unknown10"] = self.le.receiveNumber(True)

            thisEntry["unknown11"] = self.le.receiveQuotedString(True)
            thisEntry["unknown12"] = self.le.receiveQuotedString(True)
            thisEntry["unknown13"] = self.le.receiveQuotedString(True)
            thisEntry["unknown14"] = self.le.receiveNumber(True)
            thisEntry["unknown15"] = self.le.receiveNumber(True)
            thisEntry["unknown16"] = self.le.receiveQuotedString()
            resultList.append(thisEntry)
        
        self.le.receiveOK()

        return resultList

    def readFileList(self, path):
       
        self.pm.ensureEchoOff()        
        (command, heardYou)= MessageBuilder.buildReadItemsCommand(path, 'Abort')
        self.pc.write(command)

        resultList = []
        while not self.le.nextTokenIsOK():
            thisEntry = {}
            self.le.receiveConstantWord(heardYou)
            thisEntry["unknown1"] = self.le.receiveNumber(True)
            thisEntry["fileName"] = self.le.receiveQuotedString(True)
            thisEntry["unknown2"] = self.le.receiveNumber(True)

            thisEntry["unknown3"] = self.le.receiveQuotedString(True)
            thisEntry["unknown4"] = self.le.receiveNumber(True)
            thisEntry["unknown5"] = self.le.receiveNumber(True)
            thisEntry["fileSize"] = self.le.receiveNumber(True)
            thisEntry["unknown6"] = self.le.receiveNumber(True)
            thisEntry["date"] = self.le.receiveQuotedString(True)
            thisEntry["unknown7"] = self.le.receiveQuotedString(True)
            thisEntry["unknown8"] = self.le.receiveQuotedString(False)
            resultList.append(thisEntry)
        
        self.le.receiveOK()

        return resultList

    #---- Clock

    def getTime(self):
        self.pm.ensureEchoOff()        
        (command, heardYou)= MessageBuilder.buildGeneralCommand('RequestTime')
        
        self.pc.write(command)
        self.le.receiveConstantWord(heardYou)
        timeString = self.le.receiveQuotedString(False)
        #Format is: %y/%m/%d,%H:%M:%S-XX
        # Where XX is 28 for me, and can be 00-47
        # Turns out to mean GMT-offset in qrtr hours. so +/- 0-47(48?) = GMT+/-11.75
        # -28 = -7, which is weird.
        self.le.receiveOK()

        return timeString

    def setTime(self, timeString = None):
        self.pm.ensureEchoOff()
        if timeString == None:
            # Arguably this is too low a layer to do this, but it seems like such an essential command to leave out.
            # Todo - make it timezone correct. !@#$
            # Nah, that is platform specific. Move it up a layer.
            # win32all has win32timezone.py
            timeString = datetime.datetime.now().strftime("%y/%m/%d,%H:%M:%S-29")
            
        (command, heardYou)= MessageBuilder.buildGeneralCommand('SetTime', timeString=timeString)

        print command
        
        self.pc.write(command)
        assert(heardYou == '')
        self.le.receiveOK()
    
