from Lexeme import *
import Lexer

class ParseError(Exception):
    # Call it a parse error, even though it occurred in a "lexer" layer, because
    # that is what it better represents.
    
    def __init__(self, receivedToken, expectedDescription):
        self.receivedToken = receivedToken
        self.expectedDescription = expectedDescription

    def __str__(self):
        return "{ParseError: Got "+str(self.receivedToken) + ", Expected "+ self.expectedDescription+"}"

class LexerExpect:
    """Adds an extra layer over the lexer.
       Allows the Parser to be written more clearly, by expecting particular lexemes, and having
       ParserErrors raised if it isn't present.

       Lookahead utilities are also offered."""
       
    def __init__(self, lexer):
        self.lexer = lexer

    # The Receive methods will consume an expected token, returning the (text/numeric) value of the token
    # where it has one. If it is not present, a ParseError exception will be raised.
    #
    # The ReceiveConstant methods take an expected (text/numeric) value for the text, and ensure that the
    # lexeme has that value.
    
    def receiveWhitespace(self):
        self.receiveFixedLexeme(LexemeWhitespace,skipLeadingWhitespace = False)

    def receiveOK(self):
        self.receiveFixedLexeme(LexemeOK)

    def receiveError(self):
        self.receiveFixedLexeme(LexemeError)

    def receiveComma(self):
        self.receiveFixedLexeme(LexemeComma)        

    def receiveNumber(self, trailingComma = False):
        """A number is a special case of a Word lexeme that is all digits."""

        text = self.receiveFixedLexeme(LexemeWord, trailingComma)
        if text.isdigit():
            return int(text)
        else:
            raise ParseError(token,'Number')

    def receiveWord(self, trailingComma = False):
        return self.receiveFixedLexeme(LexemeWord, trailingComma)

    def receiveConstantWord(self, text, trailingComma = False):
        receivedtext = self.receiveFixedLexeme(LexemeWord, trailingComma)
        if receivedtext == text:
            return
        else:
            raise ParseError(receivedtext,'Constant word: '+text)

    def receiveQuotedString(self, trailingComma = False):
        return self.receiveFixedLexeme(LexemeQuotedString, trailingComma)        

    def receiveConstantQuotedString(self, text, trailingComma = False):
        receivedtext = self.receiveFixedLexeme(LexemeQuotedString, trailingComma)
        if receivedtext == text:
            return
        else:
            raise ParseError(token,'Constant quoted string: '+text)

    def receiveSTXQuotedString(self,trailingComma = False):
        return self.receiveFixedLexeme(LexemeSTXQuotedString, trailingComma)

    def receivePascalString(self, trailingComma = False):
        """A Pascal String is one where there is a number, representing the length
           followed by the string of that length. So-called because the Pascal language 
           used this method in parameter passing."""
        length = self.receiveNumber(False)
        self.receiveWhitespace()
        self.receiveWhitespace() # Not happy with counting chars in CR/LF solution       
        text = self.lexer.readPascalStringPart2(length).text
        if len(text) <> length:
            raise ParseError(text,'Pascal String of length:'+str(length))            
        if trailingComma:
            self.receiveComma()
        return text

    # Private implementation
    def receiveFixedLexeme(self, lexemeType, trailingComma = False, skipLeadingWhitespace = True):
        token = self.lexer.readToken(skipLeadingWhitespace)
        if isinstance(token, lexemeType):
            if trailingComma:
                self.receiveComma()
            return token.text
        else:
            raise ParseError(token,str(lexemeType))

    #----------------------------------------
    """Lookahead methods to allow the parser to branch depending on which token is coming next."""

    def isNextToken(self, lexemeType, skipLeadingWhitespace = False):
        token = self.lexer.readToken(skipLeadingWhitespace)
        self.lexer.unreadToken(token)
        return isinstance(token, lexemeType)

    def nextTokenIsOK(self):
        return self.isNextToken(LexemeOK, True)

    def nextTokenIsError(self):
        return self.isNextToken(LexemeError, True)

    def nextTokenIsWord(self):
        return self.isNextToken(LexemeWord, True)

    def nextTokenIsComma(self):
        return self.isNextToken(LexemeComma, True)

    def nextTokenIsQuotedString(self):
        return self.isNextToken(LexemeQuotedString, True)
        
