"""Create messages to send to D600 Phone.

    Has no concept of phone's state."""

# For some commands, the phone will repeat back part of the command as a prefix to the result. I called this a
# HeardYou response, because it shows the phone HeardYou correctly

# The D600 has several different ModeTypes. For each ModeType, it stores a Mode.
# E.g. the PhoneBook has one mode. The messageBook has another mode.

echoModeDefinitions = {
    'On' : '1', # Warning - Turning On Echo is likely to break response parsing elsewhere in the system.
    'Off' : '0' 
    }
phoneBookModeDefinitions = {
    'UnknownDC'             : 'DC',
    'EmergencyNumbers'      : 'EN',
    'UnknownFD'             : 'FD',
    'UnknownMC'             : 'MC',
    'UnknownON'             : 'ON',
    'UnknownRC'             : 'RC',
    'SIMCardContacts'       : 'SM',
    'PhoneMemoryContacts'   : 'ME'
    }

legalPhoneBookModes = phoneBookModeDefinitions.keys()

messageBookModeDefinitions = {    
    'PhoneMemoryMessages'   : 'ME',
    'SIMCardMessage'        : 'SM'
    }

legalMessageBookModes = messageBookModeDefinitions.keys()

modeDefinitions = {
    # Dictionary of ModeTypes.
    # Each ModeType has a description => (command, heardYou, modeDefinition)
    'Echo' : ('ATE%(ModeString)s','ATE%(ModeString)s', echoModeDefinitions),
    'PhoneBook' : ('AT+CPBS="%(ModeString)s"','',phoneBookModeDefinitions),
    'MessageBook' : ('AT+CPMS="%(ModeString)s"','+CPMS:', messageBookModeDefinitions),
    }

legalModeTypes = modeDefinitions.keys()

def buildChangeModeCommand(modeType, mode):
    assert modeType in legalModeTypes
    commandTemplate, heardYouTemplate, modeDefinition = modeDefinitions[modeType]
    assert modeDefinition.has_key(mode), str(mode) + " is not a valid mode in "+str(modeType)
    return (commandTemplate%{'ModeString': modeDefinition[mode]}+'\r', heardYouTemplate%{'ModeString': modeDefinition[mode]})
    
generalCommandDefinitions = {
    # 'CommandDescription' : ('ATCode?=%(word)', 'HeardYouResponse', parameterCount),
     'RequestPhoneBookSummary': ('AT+CPBS?', '+CPBS:', 0),
     'RequestMessageBookSummary': ('AT+CPMS?', '+CPMS:', 0),
     'RequestAppointmentBookSummary': ('AT+ORGI?', '+ORGI:',0),
     'RequestContact': ('AT+CPBR=%(position)d', '+CPBR:', 1),
     'RequestMessage': ('AT+CMGR=%(position)d', '+CMGR:', 1),
     'RequestEvent': ('AT+ORGR=%(position)d', '+ORGR:',1),
     'RequestTime': ('AT+CCLK?', '+CCLK:',0),
     'SetTime': ('AT+CCLK="%(timeString)s"', '',1),     
    }

legalCommands = generalCommandDefinitions.keys()


def buildGeneralCommand(command, **kwargs):
    assert command in legalCommands, "Don't understand command "+str(command)
    commandCode, heardYouResponse, parameterCount = generalCommandDefinitions[command%kwargs]
    assert len(kwargs) == parameterCount
    return ((commandCode%kwargs)+'\r', heardYouResponse%kwargs)


enforcementStrategies = set([
    'Force',  # If there is a value out of range, send it to the phone anyway. Useful for debugging field meanings.
    'Abort',  # Throw an exception if out of range
    'MakeDo', # Ignore fields that can't be sent. Return None if the whole command cannot be made to work.
    ])

class FieldValueError(Exception):
    def __init__(self, field, value):
        self.field = field
        self.value = value

    def __str__(self):
        return "FieldValueError: "+str(self.field)+" has bad value: "+str(self.value)
    
def strRangedNumber(fieldName, value, lowerBound, upperBound, default, enforcementStrategy):
    try:
        if value == None:
            if default <> None:
                # If there is a default, fine, use that.
                value = default
            # Otherwise
            else:
                raise ValueError
        if value <> '': # Special case of being left empty.
            intValue = int(value)
            if intValue < lowerBound or intValue > upperBound:
                raise ValueError

        return str(value)
    
    except (ValueError, TypeError),e:
        if enforcementStrategy == 'Abort':
            print fieldName, value, lowerBound, upperBound, default
            raise FieldValueError(fieldName, value)
        elif enforcementStrategy == 'MakeDo':
            if default:
                return str(default)
            else:
                return '0' # Have no better default
        elif enforcementStrategy == 'Force':
            return str(value)

def strQuotedString(fieldName, value, maxLength, default, enforcementStrategy):
    # Error-checking to be implemented
    if value == None:
        value = default
    value = value[:maxLength] # Truncate - should check instead.
    result = '"'+str(value)+'"'
    return result
        
def strQuotedString(fieldName, value, maxLength, default, enforcementStrategy):
    # Error-checking to be implemented
    if value == None:
        value = default
        assert(default <> None)
    value = value[:maxLength] # Truncate - should check instead.
    result = '"'+str(value)+'"'
    return result

STX = '\02'
ETX = '\03'

def strSTXQuotedString(fieldName, value, maxLength, default, enforcementStrategy):
    # Error-checking to be implemented
    if value == None:
        value = default
    value = value[:maxLength] # Truncate - should check instead.
    result = '"'+STX+str(value)+ETX+'"'
    return result

def strPhoneNumber(fieldName, value, default, enforcementStrategy):
    # To be implemented
    # Suffice to use quotedString for now.
    return strQuotedString(fieldName, value, 40, default, enforcementStrategy)

        
# MUST BE in PhoneMemory mode first
def buildPhoneMemoryContactWriteCommand(fields, enforcementStrategy = 'Abort'):
    assert(enforcementStrategy in enforcementStrategies)
    assert isinstance(fields,dict)    
    if enforcementStrategy <> 'Abort':
        raise NotYetImplemented
    
    result = 'AT+CPBW='

    field = strRangedNumber('contactId', fields.get('contactId'),1,65535,65535,enforcementStrategy)
    result += field+","
    
    field = strPhoneNumber('cellNumber', fields.get('cellNumber'), "", enforcementStrategy)
    result += field+","
    result += strRangedNumber("CellNumberAddressType", fields.get('cellNumberAddressType'), 129, 255, "", enforcementStrategy)
    result += field+","
    field = strSTXQuotedString('lastName', fields.get('lastName'),20,"",enforcementStrategy)
    result += field+","
    field = strSTXQuotedString('firstName', fields.get('firstName'),20,"",enforcementStrategy)
    result += field+","
    
    
    field = strPhoneNumber('homeNumber', fields.get('homeNumber'), "", enforcementStrategy)
    result += field+","
    result += "129," # secret field? !@#$

    field = strPhoneNumber('officeNumber', fields.get('officeNumber'), "", enforcementStrategy)
    result += field+","
    result += "129," # secret field? !@#$


    field = strPhoneNumber('faxNumber', fields.get('faxNumber'), "", enforcementStrategy)
    result += field+","
    result += "129," # secret field? !@#$

    field = strPhoneNumber('otherNumber', fields.get('otherNumber'), "", enforcementStrategy)
    result += field+","
    result += "129," # secret field? !@#$

    field = strQuotedString('emailAddress', fields.get('emailAddress'), 51, "", enforcementStrategy) # Or 50?
    result += field+","

    result += '"",' # Unknown field !@#$

    field = strSTXQuotedString('notes', fields.get('notes'),50,"",enforcementStrategy) # or 51?
    result += field+","

    field = strRangedNumber('groupsBitfield', fields.get('groupsBitfield'),0,2**10-1,0,enforcementStrategy)
    result += field+","
    
    field = strRangedNumber('specialRingToneTypeId', fields.get('specialRingToneTypeId'),0,65535,65535,enforcementStrategy)
    result += field+","
    
    field = strRangedNumber('builtInRingToneId', fields.get('builtInRingToneId'),0,255,255,enforcementStrategy)
    result += field+","

    field = strRangedNumber('builtInAvatarId', fields.get('builtInAvatarId'),0,255,255,enforcementStrategy)
    result += field+","

    field = strRangedNumber('specialAvatarTypeId', fields.get('specialAvatarTypeId'),0,65535,65535,enforcementStrategy)
    result += field+","

    field = strQuotedString('specialRingToneFilename', fields.get('specialRingToneFilename'), 255, "", enforcementStrategy) # Dunno max length!
    result += field+","

    field = strQuotedString('specialAvatarFilename', fields.get('specialAvatarFilename'), 255, "", enforcementStrategy) # Dunno max length!
    result += field+"\r"

    return (result, '+CPBW:')

# MUST BE in SIMCard mode first
def buildSIMCardContactWriteCommand(fields, enforcementStrategy = 'Abort'):
    assert(enforcementStrategy in enforcementStrategies)
    assert isinstance(fields,dict)    
    if enforcementStrategy <> 'Abort':
        raise NotYetImplemented
    
    result = 'AT+CPBW='

    field = strRangedNumber('contactId', fields.get('contactId'),1,200,None,enforcementStrategy)
    result += field+","
    
    field = strPhoneNumber('phoneNumber', fields.get('phoneNumber'), None, enforcementStrategy)
    result += field+","
    result += "129," # Mystery field
    field = strSTXQuotedString('name', fields.get('name'),16,"",enforcementStrategy)
    result += field+"\r"
    
    return (result, '+CPBW:')

# Be sure to set the mode as desired first.
def buildEventWriteCommand(fields, enforcementStrategy = 'Abort'):
    assert(enforcementStrategy in enforcementStrategies)
    assert isinstance(fields,dict)    
    if enforcementStrategy <> 'Abort':
        raise NotYetImplemented

    # To Do: Enforce relationships between dates (e.g. endDate > startDate)
    
    result = 'AT+ORGW='

    eventType = fields.get('eventType')

    SCHEDULE = 1
    ANNIVERSARY = 2
    TODO = 3
    MISCELLANEOUS = 4

    maxAlarmBeforeCountByUnits = {
        0 :  0, # Makes no sense otherwise.
        1 : 59, # 59 minutes
        2 : 23, # 23 hours
        3 : 30, # 30 days
        4 :  2, # 2 weeks
        }

    field = strRangedNumber('eventId', fields.get('eventId'),0,65535,65535,enforcementStrategy)
    result += field+","

    result += "0," # Don't know why a leading number, that doesn't appear in the read op.
    # This unknown particularly concerns me.
    
    field = strRangedNumber('eventType', eventType,1,4,None,enforcementStrategy)
    result += field+","

    if eventType in [SCHEDULE, MISCELLANEOUS]:
        field = strSTXQuotedString('subject', fields.get('subject'),15,"",enforcementStrategy)
        result += field+","
    else:
        result += ","

    field = strSTXQuotedString('text', fields.get('text'),100,"",enforcementStrategy)
    result += field+","

    field = strRangedNumber('startDay', fields.get('startDay'),1,31,None,enforcementStrategy)
    result += field+","
    field = strRangedNumber('startMonth', fields.get('startMonth'),1,12,None,enforcementStrategy)
    result += field+","
    field = strRangedNumber('startYear', fields.get('startYear'),2000,2098,None,enforcementStrategy)
    result += field+","

    if eventType in [SCHEDULE, MISCELLANEOUS, ANNIVERSARY]:
        field = strRangedNumber('startHour', fields.get('startHour'),0,23,None,enforcementStrategy)
        result += field+","
        field = strRangedNumber('startMinute', fields.get('startMinute'),0,59,None,enforcementStrategy)
        result += field+","
    else:
        result += ",,"

    if eventType in [SCHEDULE, MISCELLANEOUS, TODO]:
        field = strRangedNumber('endDay', fields.get('endDay'),1,31,None,enforcementStrategy)
        result += field+","
        field = strRangedNumber('endMonth', fields.get('endMonth'),1,12,None,enforcementStrategy)
        result += field+","
        field = strRangedNumber('endYear', fields.get('endYear'),2000,2098,None,enforcementStrategy)
        result += field+","
    else:
        result += ",,,"
        
    if eventType in [SCHEDULE, MISCELLANEOUS]:
        field = strRangedNumber('endHour', fields.get('endHour'),0,23,None,enforcementStrategy)
        result += field+","
        field = strRangedNumber('endMinute', fields.get('endMinute'),0,59,None,enforcementStrategy)
        result += field+","
    else:
        result += ",,"
        
    if eventType in [SCHEDULE]:        
        field = strSTXQuotedString('location', fields.get('location'),15,"",enforcementStrategy)
        result += field+","
    else:
        result += ","

    if eventType == ANNIVERSARY:
        # Can only have days or weeks notice.
        # Should use enforcement rules to decide action.
        if fields.get('alarmBeforeUnits',0) not in [0,3,4]:
            raise FieldValueError('alarmBeforeUnits',fields.get('alarmBeforeUnits',0))

    if eventType in [SCHEDULE, ANNIVERSARY]:
        field = strRangedNumber('alarmOn', fields.get('alarmOn'),0,1,0,enforcementStrategy)
        result += field+","
        
        field = strRangedNumber('alarmBeforeUnits', fields.get('alarmBeforeUnits'),0,4,0,enforcementStrategy)
        result += field+","
        field = strRangedNumber('alarmBeforeCount', fields.get('alarmBeforeCount'),0,maxAlarmBeforeCountByUnits[fields.get('alarmBeforeUnits',0)],0,enforcementStrategy)
        result += field+","
    else:
        result += ",,,"

    if eventType in [SCHEDULE, ANNIVERSARY]:
        field = strRangedNumber('recurrence', fields.get('recurrence'),0,4,0,enforcementStrategy)        
        result += field+","
    else:
        result += ","

    if eventType in [TODO]:
        field = strRangedNumber('priority', fields.get('priority'),1,3,2,enforcementStrategy)
        result += field+","
        field = strRangedNumber('complete', fields.get('complete'),0,1,1,enforcementStrategy)
        result += field+","
    else:
        result += ",,"

    if eventType in [SCHEDULE] and fields.get('recurrence',0)>0:
        field = strRangedNumber('recurrenceEndDay', fields.get('recurrenceEndDay'),1,31,None,enforcementStrategy)
        result += field+","
        field = strRangedNumber('recurrenceEndMonth', fields.get('recurrenceEndMonth'),1,12,None,enforcementStrategy)
        result += field+","
        field = strRangedNumber('recurrenceEndYear', fields.get('recurrenceEndYear'),2000,2098,None,enforcementStrategy)
        result += field # No comma
    else:
        result += ",,"

    result +="\r"

    return (result, '+ORGW:')
    

# MUST BE in PhoneMemory mode first
def buildPhoneMemoryContactDeleteCommand(contactId, enforcementStrategy = 'Abort'):
    assert(enforcementStrategy in enforcementStrategies)
    
    result = 'AT+CPBW='

    field = strRangedNumber('contactId', contactId,0,65534,None,enforcementStrategy)
    result += field+"\r"

    return (result, '+CPBW:')

# MUST BE in SIMCard mode first
def buildSIMCardContactDeleteCommand(contactId, enforcementStrategy = 'Abort'):
    assert(enforcementStrategy in enforcementStrategies)
    
    result = 'AT+CPBW='

    field = strRangedNumber('contactId', contactId,1,200,None,enforcementStrategy)
    result += field+"\r"

    return (result, '+CPBW:')

def buildEventDeleteCommand(eventId, enforcementStrategy = 'Abort'):
    assert(enforcementStrategy in enforcementStrategies)
    
    result = 'AT+ORGD='

    field = strRangedNumber('eventId', eventId,0,65534,None,enforcementStrategy)
    result += field+"\r"

    return (result, None)

def buildMessageDeleteCommand(messageId, enforcementStrategy = 'Abort'):
    assert(enforcementStrategy in enforcementStrategies)
    
    result = 'AT+CMGD='

    field = strRangedNumber('messageId', messageId,0,200,None,enforcementStrategy)
    result += field+"\r"

    return (result, None)


def buildReadFoldersCommand(path, enforcementStrategy = 'Abort'):

    assert(enforcementStrategy in enforcementStrategies)
    
    result = 'AT+FSDI='

    field = strQuotedString("path",path, 65535,"/",enforcementStrategy)
    result += field+"\r"

    return (result, '+FSDI:')

def buildReadItemsCommand(path, enforcementStrategy = 'Abort'):

    assert(enforcementStrategy in enforcementStrategies)
    
    result = 'AT+FSDL='

    field = strQuotedString("path", path, 65535,"/",enforcementStrategy)
    result += field+"\r"

    return (result, '+FSDL:')
    

#-------------------------------------------------

if __name__ == '__main__':
    assert buildChangeModeCommand('Echo', 'Off') == ('ATE0\r', 'ATE0')
    assert buildChangeModeCommand('PhoneBook', 'SIMCardContacts') == ('AT+CPBS="SM"\r','')
    assert buildGeneralCommand('RequestPhoneBookSummary') == ('AT+CPBS?\r', '+CPBS:')
    assert buildGeneralCommand('RequestMessage',position = 43) == ('AT+CMGR=43\r', '+CMGR:')
    print "Unit Tests Passed"

    
