import CommandProtocol
import collections
import datetime
from D600ParserLog import *

def cleanNumber(numberString):
    if numberString == None:
        return None
    result = ''
    for ch in str(numberString):
        if ch in "0123456789+*#P":
            result += ch
    return result

class PhoneItemLocation():
    validItemKinds = set(['Contact_SIM','Contact_PhoneMemory','Event'])
#    validLocationKind = set(['read','create','delete'])

    # This should be protected for multi-threaded ops.
    newItem = 65536

    def __init__(self,itemKind,readItemNumber = None, createItemNumber = None, deleteItemNumber = None):
        assert itemKind in PhoneItemLocation.validItemKinds
        self.itemKind = itemKind
        if createItemNumber == None:
            createItemNumber = PhoneItemLocation.newItem
            PhoneItemLocation.newItem += 1
        self.itemNumbers = (readItemNumber, createItemNumber, deleteItemNumber)

    def isItemKind(self,itemKind):
        return self.itemKind == itemKind

    def readItemNumber(self):
        return self.itemNumbers[0]
        
    def createItemNumber(self):
        createItemNumber = self.itemNumbers[1]
        if createItemNumber > 65535:
            createItemNumber = 65535
        return createItemNumber

    def deleteItemNumber(self):
        return self.itemNumbers[2]

    def __str__(self):
        return str(self.itemKind) + str(self.itemNumbers)
    def __hash__(self):
        return hash((self.itemKind, self.itemNumbers))

    
class PhoneItem():
    # Protected
    def __init__(self, phoneFields, cp):
        raise NotImplementedError 

    #Public
    def write(self, cp):
        raise NotImplementedError 

    def delete(self, cp):
        raise NotImplementedError 

    def isEquivalent(self, other):
        """Compares only significant fields to find if the two objects are equivalent"""
        raise NotImplementedError 

    def fullLocation(self):
        return self.location

    @staticmethod
    def readFactory(location, cp):
        # Warning: The location you pass to the readFactory may well be different to the objects claimed location.
        # That's because objects some objects have a READ address versus an OVERWRITE/DELETE address.
        if location.isItemKind('Contact_SIM'):
            return Contact_SIM.readFactory(location,cp)
        elif location.isItemKind('Contact_PhoneMemory'):
            return Contact_PhoneMemory.readFactory(location,cp)
        elif location.isItemKind('Event'):
            return Event.readFactory(location,cp)
        else:
            raise NotImplementedError 

class Contact(PhoneItem):
    pass

class Contact_SIM(PhoneItem):
    def __init__(self, position, name, phoneNumber):
        assert isinstance(position, int)
        self.location = PhoneItemLocation('Contact_SIM', position,position,position)
        self.phoneNumber = phoneNumber
        # Ignore AddressType
        self.name = name

    def isEquivalent(self, other):
        if (isinstance(other, Contact_SIM)
            and self.phoneNumber == other.phoneNumber
            and self.name == other.name):
            return True
        else:
            return False

    def __str__(self):
        return "[Contact_SIM: "+str(self.location)+", "+str(self.name)+"]"

    @staticmethod
    def readFactory(location, cp):
        assert location.isItemKind('Contact_SIM')
        itemNumber = location.readItemNumber()
        assert isinstance(itemNumber, int)
        phoneFields = cp.getSIMMemoryContact(itemNumber)
        if not phoneFields:
            return None
        assert itemNumber == phoneFields['locationInSIM']        
        name = phoneFields.get('name')
        phoneNumber = phoneFields.get('phoneNumber')
        return Contact_SIM(itemNumber, name, phoneNumber)



class Contact_PhoneMemory(PhoneItem):
    def __init__(self,  
            contactId = None, # The one used for update and delete, not read.
            firstName = None,
            lastName = None,
            cellNumber = None,                 
            homeNumber = None,
            officeNumber = None,
            faxNumber = None,
            otherNumber = None,
            emailAddress = None,
            notes = None,
            groupsBitfield = None,
            specialRingToneTypeId = None,
            builtInRingToneId = None,
            builtInAvatarId = None,
            specialAvatarTypeId = None,
            specialRingToneFilename = None,
            specialAvatarFilename = None
                 ):

        assert contactId is None or isinstance(contactId, int)
        self.location = PhoneItemLocation(
            'Contact_PhoneMemory',
            readItemNumber = None, # Once it is read, you can't be sure where it will be anymore.
            createItemNumber = None, # We don't have a set position to create it at.
            deleteItemNumber = contactId
            )
        self.cellNumber = cleanNumber(cellNumber)
        self.lastName = lastName
        self.firstName = firstName
        self.homeNumber = cleanNumber(homeNumber)
        self.officeNumber = cleanNumber(officeNumber)
        self.faxNumber = cleanNumber(faxNumber)
        self.otherNumber = cleanNumber(otherNumber)
        self.emailAddress = emailAddress
        self.notes = notes
        self.groupsBitfield = groupsBitfield
        self.specialRingToneTypeId = specialRingToneTypeId
        self.builtInRingToneId = builtInRingToneId
        self.builtInAvatarId = builtInAvatarId
        self.specialAvatarTypeId = specialAvatarTypeId
        self.specialRingToneFilename = specialRingToneFilename
        self.specialAvatarFilename = specialAvatarFilename

    def write(self, cp):
        fields = {
            'contactId' : self.location.createItemNumber(),
            'cellNumber' : self.cellNumber,
            'lastName' : None if self.lastName == None else self.lastName.encode('utf-8'), #,'ignore'),
            'firstName' : None if self.firstName == None else self.firstName.encode('utf-8'), #,'ignore'),
            'homeNumber' : self.homeNumber,
            'officeNumber' : self.officeNumber,
            'faxNumber' : self.faxNumber,
            'otherNumber' : self.otherNumber,
            'emailAddress' : self.emailAddress,
            'notes' : self.notes,
            'groupsBitfield' : self.groupsBitfield, 
            'specialRingToneTypeId' : self.specialRingToneTypeId,             
            'builtInRingToneId' : self.builtInRingToneId, 
            'builtInAvatarId' : self.builtInAvatarId,             
            'specialAvatarTypeId' : self.specialAvatarTypeId, 
            'specialRingToneFilename' : self.specialRingToneFilename,             
            'specialRingToneFilename' : self.specialRingToneFilename, 
            'specialAvatarFilename' : self.specialAvatarFilename
            }             
        cp.writePhoneMemoryContact(fields)
            
    def delete(self, cp):
        cp.deletePhoneMemoryContact(self.location.deleteItemNumber())

    def isEquivalent(self, other):
        return (isinstance(other, Contact_PhoneMemory)
            and self.cellNumber == other.cellNumber
            and self.lastName == other.lastName
            and self.firstName == other.firstName
            and self.homeNumber == other.homeNumber 
            and self.officeNumber == other.officeNumber
            and self.faxNumber == other.faxNumber
            and self.otherNumber == other.otherNumber
            and self.emailAddress == other.emailAddress
            and self.notes == other.notes
            #and self.groupsBitfield == other.groupsBitfield
            #and self.specialRingToneTypeId == other.specialRingToneTypeId
            #and self.builtInRingToneId == other.builtInRingToneId
            #and self.builtInAvatarId == other.builtInAvatarId
            #and self.specialAvatarTypeId == other.specialAvatarTypeId
            #and self.specialRingToneFilename == other.specialRingToneFilename
            #and self.specialAvatarFilename == other.specialAvatarFilename
            )

    def __str__(self):
        return "[Contact_PhoneMemory: "+str(self.location[1])+", "+str(self.lastName)+"]"

    @staticmethod
    def readFactory(location, cp):
        assert location.isItemKind('Contact_PhoneMemory')
        itemNumber = location.readItemNumber()
        phoneFields = cp.getPhoneMemoryContact(itemNumber)
        if not phoneFields:
            return None

        return Contact_PhoneMemory(
            contactId = phoneFields["contactId"],
            cellNumber = phoneFields["cellNumber"],
            lastName = phoneFields["lastName"].decode('utf-8'),
            firstName = phoneFields["firstName"].decode('utf-8'),
            homeNumber = phoneFields["homeNumber"],
            officeNumber = phoneFields["officeNumber"],
            faxNumber = phoneFields["faxNumber"],
            otherNumber = phoneFields["otherNumber"],
            emailAddress = phoneFields["emailAddress"],
            notes = phoneFields["notes"],
            groupsBitfield = phoneFields["groupsBitfield"],
            specialRingToneTypeId = phoneFields["specialRingToneTypeId"],
            builtInRingToneId = phoneFields["builtInRingToneId"],
            builtInAvatarId = phoneFields["builtInAvatarId"],
            specialAvatarTypeId = phoneFields["specialAvatarTypeId"],
            specialRingToneFilename = phoneFields["specialRingToneFilename"],
            specialAvatarFilename = phoneFields["specialAvatarFilename"]
                                   )
import math
class AlarmSettings:
    def __init__(self, interval, units):
        assert units in ['Immediate','Minutes','Hours','Days','Weeks']
        if units <> 'Immediate' or interval == 0:
            DOMlogger.info('Correcting bad units for alarm')
            units = 'Days'
        assert units <> 'Minutes' or (interval > 0 and interval <= 59)
        assert units <> 'Hours' or (interval > 0 and interval <= 23)
        assert units <> 'Days' or (interval > 0 and interval <= 30)
        assert units <> 'Weeks' or (interval > 0 and interval <= 2)                
                
        self.units = units
        self.interval = interval

    def __str__(self):
        return str(self.interval)+ " " + str(self.units)

    @staticmethod
    def minutesFactory(minutesBefore,daysIsMinUnit):
        # Phone can't represent arbitrary number of minutes, so it needs to be
        # approximated.
        # To Do, round UP not DOWN where safe.
        assert minutesBefore >=0
        if not daysIsMinUnit and minutesBefore <= 59:
            return AlarmSettings(minutesBefore,'Minutes')
        if not daysIsMinUnit and math.ceil(minutesBefore/60.0) < 24.0:
            return AlarmSettings(int(math.ceil(minutesBefore/60.0)),'Hours')
        if math.ceil(minutesBefore/(60.0*24)) < 31.0:
            return AlarmSettings(int(math.ceil(minutesBefore/(60.0*24))),'Days') # Days
        return AlarmSettings(min(2,int(math.ceil(minutesBefore/(60.0*24*7)))),'Weeks') # 2 Weeks is as early as we can manage.

class ScheduleRecurrence:
    validFrequencies = set(['Daily', 'Weekly', 'Monthly', 'Yearly'])
    def __init__(self,
                 frequency,
                 endDate):
        self.frequency = frequency
        self.endDate = endDate

class Event(PhoneItem):
    def __init__(self,  
            eventId = None, # The one used for update and delete, not read.
            text = None):
        self.location = PhoneItemLocation('Event', None,None,eventId)
        self.eventId = eventId
        self.text = text

    def delete(self, cp):
        cp.deleteEvent(self.location.deleteItemNumber())

    def write(self,cp):
        raise NotImplemented

    def isEquivalent(self, other):
        raise NotImplemented

    @staticmethod
    def readFactory(location, cp):
        assert location.isItemKind('Event')
        itemNumber = location.readItemNumber()
        assert isinstance(itemNumber, int)
        phoneFields = cp.getEvent(itemNumber)
        if not phoneFields:
            return None
        eventType = phoneFields["eventType"]
        assert eventType in [1,2,3,4]
        # 1 = Schedule, 2 = Anniversary, 3 = To do,4 = Miscellaneous        
        if eventType == 1: # Schedule
            return Schedule.fieldsFactory(phoneFields)
        elif eventType == 2:
            return Anniversary.fieldsFactory(phoneFields)
        elif eventType == 3:
            return ToDo.fieldsFactory(phoneFields)
        else:
            return Miscellaneous.fieldsFactory(phoneFields)

class Anniversary(Event):
    def __init__(self,
                 eventId,
                 text,
                 startDateTime,
                 alarmSettings = None,
                 annualRecurrence = None
                 ):
        Event.__init__(self,eventId,text)
        assert(self.text == text)
        self.startDateTime = startDateTime
        self.alarmSettings = alarmSettings
        DOMlogger.info("AlarmSettings: " + str(alarmSettings))
        self.annualRecurrence = annualRecurrence

    def isEquivalent(self, other):
        return (isinstance(other, Anniversary)
                and self.text == other.text
                and self.startDateTime == other.startDateTime
                and self.alarmSettings == other.alarmSettings
                and self.annualRecurrence == other.annualRecurrence
                )

    def write(self, cp):
        fields = {
            'eventId' : self.location.createItemNumber(),
            'eventType' : 2,
            'text' : self.text,
            'startDay' : self.startDateTime.day,
            'startMonth' : self.startDateTime.month,
            'startYear' : self.startDateTime.year,
            'startHour' : self.startDateTime.hour,
            'startMinute' : self.startDateTime.minute,
            'alarmOn' : 1 if self.alarmSettings else 0,
            'alarmBeforeUnits' : 0 if not self.alarmSettings else (
                {'Weekly' : 4, 'Days' : 3}[self.alarmSettings.units]
                ),
            'alarmBeforeCount' : 0 if not self.alarmSettings else self.alarmSettings.interval,
            'recurrence' : 4 if self.annualRecurrence else 0,
            }             
        cp.writeEvent(fields)
        
    @staticmethod
    def fieldsFactory(phoneFields):
        itemNumber = phoneFields.get('eventId')
        text = phoneFields.get('text')
        startDateTime = datetime.datetime(
            year = phoneFields.get("startYear"),
            month = phoneFields.get("startMonth"),
            day = phoneFields.get("startDay"),
            hour = phoneFields.get("startHour"),
            minute = phoneFields.get("startMinute")
            )
        alarmSettings = None
        if phoneFields.get("alarmOn",0) == 1:
            units = {0 : 'Immediate',
                     1: 'Minutes',
                     2: 'Hours',
                     3: 'Days',
                     4: 'Weeks'}[phoneFields.get("alarmUnits",0)]
            interval = phoneFields.get("alarmBeforeCount",0)
            alarmSettings = AlarmSettings(interval, units)
        annualRecurrence = phoneFields.get('recurrence', 0) == 4
        a = Anniversary(
            itemNumber,
            text,
            startDateTime,
            alarmSettings,
            annualRecurrence
            )
        
        return a

class Miscellaneous(Event):
    def __init__(self,
                 eventId,
                 text,
                 startDateTime,
                 subject,
                 endDate):
        Event.__init__(self,eventId,text)
        self.startDateTime = startDateTime
        self.subject = subject
        self.endDate = endDate

    def isEquivalent(self, other):
        return (isinstance(other, Miscellaneous)
                and self.text == other.text
                and self.startDateTime == other.startDateTime
                and self.subject == other.subject
                and self.endDate == other.endDate
                )

    @staticmethod
    def fieldsFactory(phoneFields):
        itemNumber = phoneFields.get('eventId')
        text = phoneFields.get('text')
        startDateTime = datetime.datetime(
            year = phoneFields.get("startYear"),
            month = phoneFields.get("startMonth"),
            day = phoneFields.get("startDay"),
            hour = phoneFields.get("startHour"),
            minute = phoneFields.get("startMinute")
            )
        subject = phoneFields.get('subject')
        endDate = datetime.date(
            year = phoneFields.get("endYear"),
            month = phoneFields.get("endMonth"),
            day = phoneFields.get("endDay")
            )

        return Miscellaneous(
            itemNumber,
            text,
            startDateTime,
            subject,
            endDate
            )
            
class Schedule(Miscellaneous):
    def __init__(self,
                 itemNumber,
                 text,
                 startDateTime,
                 subject,
                 endDate,
                 eventLocation = None,
                 alarmSettings = None,
                 recurrenceSettings = None):
        Miscellaneous.__init__(
            self,
            itemNumber,
            text,
            startDateTime,
            subject,
            endDate)
        self.eventLocation = eventLocation
        self.alarmSettings = alarmSettings
        self.recurrenceSettings = recurrenceSettings

    def isEquivalent(self, other):
        return (isinstance(other, Miscellaneous)
                and self.text == other.text
                and self.startDateTime == other.startDateTime
                and self.subject == other.subject
                and self.endDate == other.endDate
                and self.eventLocation == other.eventLocation
                and self.alarmSettings == other.alarmSettings
                and self.recurrenceSettings == other.recurrenceSettings
                )

    def write(self, cp):
        fields = {
            'eventId' : self.location.createItemNumber(),
            'eventType' : 1,
            'text' : self.text,
            'startDay' : self.startDateTime.day,
            'startMonth' : self.startDateTime.month,
            'startYear' : self.startDateTime.year,
            'startHour' : self.startDateTime.hour,
            'startMinute' : self.startDateTime.minute,
            'subject': self.subject,
            'endYear' : self.endDate.year,
            'endMonth' : self.endDate.month,            
            'endDay' : self.endDate.day,
            'endHour' : self.endDate.hour,
            'endMinute' : self.endDate.minute,
            'location' : self.eventLocation,
            'alarmOn' : 1 if self.alarmSettings else 0,
            'alarmBeforeUnits' : 0 if not self.alarmSettings else (
                {'Weekly' : 4, 'Days' : 3, 'Hours' : 2, 'Minutes' : 1}[self.alarmSettings.units]
                ),
            'alarmBeforeCount' : 0 if not self.alarmSettings else self.alarmSettings.interval,
            'recurrence' : 0 if not self.recurrenceSettings else (
                { "Daily" : 1,"Weekly" : 2,"Monthly" : 3, "Yearly" : 4}[self.recurrenceSettings.frequency]
                ),
            'recurrenceEndYear' : None if not self.recurrenceSettings else self.recurrenceSettings.year,
            'recurrenceEndMonth' : None if not self.recurrenceSettings else self.recurrenceSettings.month,
            'recurrenceEndDay' : None if not self.recurrenceSettings else self.recurrenceSettings.day,            
            }             
        cp.writeEvent(fields)                

    @staticmethod
    def fieldsFactory(phoneFields):
        itemNumber = phoneFields.get('eventId')
        text = phoneFields.get('text')
        startDateTime = datetime.datetime(
            year = phoneFields.get("startYear"),
            month = phoneFields.get("startMonth"),
            day = phoneFields.get("startDay"),
            hour = phoneFields.get("startHour"),
            minute = phoneFields.get("startMinute")
            )
        subject = phoneFields.get('subject')
        endDate = datetime.date(
            year = phoneFields.get("endYear"),
            month = phoneFields.get("endMonth"),
            day = phoneFields.get("endDay")
            )
        eventLocation = phoneFields.get("location")
        alarmSettings = None
        if phoneFields.get("alarmOn"):
            alarmSettings = AlarmSettings(interval = phoneFields.get("alarmBeforeCount"),
                                          units = { 1 : "Minutes", 2 : "Hours", 3: "Days", 4: "Weeks" }[phoneFields.get("alarmBeforeUnits")]
                                          )
        recurrenceSettings = None
        if phoneFields.get("recurrence", 0) <> 0:
            recurrenceSettings = RecurrenceSettings(
                frequency = { 1 : "Daily", 2 : "Weekly", 3: "Monthly", 4: "Yearly" }[phoneFields["recurrence"]],
                endDate = datetime.date(
                    year = phoneFields.get("recurrenceEndYear"),
                    month = phoneFields.get("recurrenceEndMonth"),
                    day = phoneFields.get("recurrenceEndDay")
                    )
                )

        return Schedule(
            itemNumber,
            text,
            startDateTime,
            subject,
            endDate,
            eventLocation,
            alarmSettings,
            recurrenceSettings)


class ToDo(Event):
    def __init__(self,
                 eventId,
                 startDate,
                 endDate,
                 text,
                 priority,
                 isComplete):
        Event.__init__(self,eventId,text)
        self.startDate = startDate
        self.endDate = endDate        
        self.priority = priority
        self.isComplete = isComplete

    def isEquivalent(self, other):
        return (isinstance(other, Miscellaneous)
                and self.text == other.text
                and self.startDate == other.startDate
                and self.endDate == other.endDate
                and self.eventLocation == other.eventLocation
                and self.priority == other.priority
                and self.isComplete == other.isComplete
                )

    def write(self, cp):
        fields = {
            'eventId' : self.location.createItemNumber(),
            'eventType' : 3,
            'text' : self.text,
            'startDay' : self.startDate.day,
            'startMonth' : self.startDate.month,
            'startYear' : self.startDate.year,
            'endDay' : self.startDate.day,
            'endMonth' : self.startDate.month,
            'endYear' : self.startDate.year,
            'priority' : {'High' : 3, 'Normal' : 2, 'Low' : 1}[self.priority],
            'complete' : 1 if self.isComplete else 2,
            }
        cp.writeEvent(fields)

    @staticmethod
    def fieldsFactory(phoneFields):
        itemNumber = phoneFields.get('eventId')
        text = phoneFields.get('text')
        startDate = datetime.date(
            phoneFields.get('startYear'),
            phoneFields.get('startMonth'),
            phoneFields.get('startDay')
            )
        endDate = datetime.date(
            phoneFields.get('endYear'),
            phoneFields.get('endMonth'),
            phoneFields.get('endDay')
            )
        priority = {3: 'High', 2: 'Normal', 1 : 'Low'}[phoneFields.get('priority')]
        isComplete = phoneFields.get('complete') == 1
        return ToDo(
                 itemNumber,
                 startDate,
                 endDate,
                 text,
                 priority,
                 isComplete)            
            
class PhoneItemDict(collections.MutableMapping):
    def __init__(self):
        self.dict = {}

    def __getitem__(self,b):
        return self.dict.__getitem__(b)
    
    def __setitem__(self,a,b):
        return self.dict.__setitem__(a,b)

    def __delitem__(self,b):
        return self.dict.__delitem__(b)        

    def __iter__(self):
        return self.dict.__iter__()

    def __len__(self):
        return len(self.dict)

    def add(self, item):
        self.dict[item.fullLocation()] = item

    def writeToPhone(self,cp,callback = None):
        total = len(self.dict.values())
        count = 0
        for item in self.dict.values():
            item.write(cp)
            count = count + 1
            if callback: 
                callback(count,total, "Write item to phone")
    
    def deleteFromPhone(self,cp, callback = None):
        total = len(self.dict.values())
        count = 1
        for item in self.dict.values():
            item.delete(cp)
            count = count + 1
            if callback: 
                callback(count,total, "Deleted item from phone")
            


def readAllPhoneMemoryContacts(ps, callback = None):
    count = ps.getPhoneBookSummary('PhoneMemoryContacts')["count"]
    result = PhoneItemDict()
    if callback:
        callback(1, count+1, "Read phonebook summary")
    
    for i in range(1,count+1):
        c = PhoneItem.readFactory(
                PhoneItemLocation('Contact_PhoneMemory',readItemNumber = i),
                ps)
        result.add(c)
        if callback:
            summary = "Read " + (c.firstName + " " if c.firstName <> None else "") + (c.lastName if c.lastName <> None else "") + " contact"
            callback(i+1, count+1, summary)
    return result

def readAllEvents(ps, callback = None):
    count = ps.getAppointmentBookSummary()["count"]
    result = PhoneItemDict()
    if callback:
        callback(1, count+1, "Read appointment book summary")
    
    for i in range(0,count):
        c = PhoneItem.readFactory(
                PhoneItemLocation('Event',readItemNumber = i),
                ps)
        result.add(c)
        if callback:
            callback(i+2, count+1, "Read event")
    return result


#-------------------------------------------------

if __name__ == '__main__':

    l = PhoneItemLocation('Contact_SIM',1,None,3)
    assert l.isItemKind('Contact_SIM')
    assert l.readItemNumber() == 1
    print l.createItemNumber()
    assert l.createItemNumber() == 65535
    assert l.deleteItemNumber() == 3
    

    pid = PhoneItemDict()
    pid[1] = 'a'
    pid[2] = 'b'
    for i in pid:
        print i,pid[i]
    print len(pid)
    del pid[1]
    print len(pid)    
        
    
    import PhoneConnection
    
    pc = PhoneConnection.SerialPhoneConnection()
    try:
        cp = CommandProtocol.CommandProtocol(pc)

        a = PhoneItem.readFactory(PhoneItemLocation('Contact_SIM', readItemNumber = 3), cp)
        b = PhoneItem.readFactory(PhoneItemLocation('Contact_SIM', readItemNumber = 4), cp)
        c = PhoneItem.readFactory(PhoneItemLocation('Contact_SIM', readItemNumber = 3), cp)
        assert a.isEquivalent(c) and c.isEquivalent(a) and not b.isEquivalent(a)
        

        count = cp.getAppointmentBookSummary()["count"]
        result = PhoneItemDict()
        
        for i in range(0,count):
#            try:
                c = PhoneItem.readFactory(
                        PhoneItemLocation('Event',readItemNumber = i),
                        cp)
                if c:
                    result.add(c)
                    print c,c.text
#            except Exception, e:
#                print e,
#                raise e
        print len(result)

    finally:
        pc.close()

    print "Unit Tests Passed"
